"use strict";

/**
 * noDarkmode Frontend Core
 *
 * @author  reDim GmbH | Elias Ritter
 * @license GNU General Public License 2.0 or later, see https://www.gnu.org/licenses/old-licenses/gpl-2.0.en.html
 * @since   version 1.0
 */
class noDarkmode {

    /**
     * Toggler Class
     *
     * @type {Toggler}
     */
    static Toggler = class {

        /**
         * Class Constructor
         *
         * @param {string} classname
         */
        constructor(classname) {
            this.dom = this.createElement("a", "reDimnoDarkmodeToggler " + classname);
            this.icon_container = this.createElement("div", "header-item-icon");
            this.icon = this.createElement("span", "icon-fa");

            const menubar = document.querySelector(".header-items.d-flex.ms-auto");
            menubar.insertBefore(this.dom, menubar.children[1]);
        }

        /**
         * Create a HTMLElement with a specific class
         *
         * @param {string} element
         * @param {string} classname
         * @return {HTMLElement}
         */
        createElement(element, classname = "") {
            const domelement = document.createElement(element);

            if (classname) {
                domelement.className = classname;
            }

            return domelement;
        }

        /**
         * Set the current Toggler state
         *
         * @param {string} state
         */
        setState(state) {
            switch (state) {

                case "inactive":
                    this.icon.classList.add("fa-moon");
                    this.dom.title = "noDarkmode aktivieren";
                    this.dom.classList.add("inactive");
                    break;

                case "active":
                default:
                    this.icon.classList.add("fa-sun");
                    this.dom.title = "noDarkmode deaktivieren";
                    this.dom.classList.add("active");
                    break;
            }

            this.icon_container.appendChild(this.icon);
            this.dom.appendChild(this.icon_container);
        }

    }

    /**
     * Language Helper Class
     *
     * @type {Language}
     */
    static Language = class {

        /**
         * German Translation
         *
         * @type {Object}
         */
        lang_de_DE = {
            ND_ASSETS_DEACTIVATED: "Folgende Assets wurden deaktiviert oder modifiziert: \n\r",
            ND_NOT_BROWSER_ENABLED: "Darkmode wird nicht vom Browser gefordert, daher hat sich das noDarkmode-Plugin automatisch deaktiviert"
        };

        /**
         * English Translation
         *
         * @type {Object}
         */
        lang_en_GB = {
            ND_ASSETS_DEACTIVATED: "The following assets have been deactivated or modified: \n\r",
            ND_NOT_BROWSER_ENABLED: "Darkmode is not required by the browser, therefore the noDarkmode plugin is automatically deactivated"
        }

        /**
         * Class Constructor
         *
         * @param {string} lang
         */
        constructor(lang) {
            lang = lang.replace('-', '_');

            switch (lang) {

                case 'de_DE':
                case 'en_GB':
                    this.lang = lang;
                    break;

                case 'en_US':
                default:
                    this.lang = 'en_GB';
                    break;
            }
        }

        /**
         * Translates a string into the currently set language
         *
         * @param {string} str
         * @param {string} prefix
         * @returns {string}
         */
        _(str, prefix = "noDarkmode: ") {

            const langarr = this['lang_' + this.lang]

            for (let key in langarr) {
                if (langarr.hasOwnProperty(key)) {
                    let value = langarr[key];

                    return prefix + str.replace(key, value);
                }
            }
        }
    }

    /**
     * Get the Value of a Cookie
     *
     * @param {string} cookieName
     * @return {undefined|string}
     */
    static getCookie(cookieName) {
        const cookies = document.cookie.split(";");

        for (let i = 0; i < cookies.length; i++) {
            const cookie = cookies[i].trim();

            if (cookie.indexOf(cookieName + "=") === 0) {
                return cookie.substring(cookieName.length + 1);
            }
        }

        return void(0);
    }

    /**
     * Checks, if the Browser currently prefers a specific Color-scheme
     *
     * @returns {boolean}
     */
    static browserPrefersScheme(scheme) {
        return (window.matchMedia && window.matchMedia('(prefers-color-scheme: ' + scheme + ')').matches);
    }
}