<?php

namespace Joomla\Plugin\System\nodarkmode\Extension;

use Joomla\CMS\Factory;
use Joomla\CMS\Language\LanguageFactoryInterface;
use Joomla\CMS\Language\Text;
use Joomla\CMS\Uri\Uri;
use Joomla\CMS\WebAsset\WebAssetItem;
use Joomla\Event\DispatcherInterface;
use Joomla\Filesystem\File;
use Joomla\CMS\Plugin\CMSPlugin;
use Joomla\CMS\WebAsset\WebAssetManager;
use Joomla\CMS\Application\CMSApplicationInterface;

defined('_JEXEC') or die;

/**
 * noDarkmode Plugin
 *
 * @author  reDim GmbH | Elias Ritter
 * @license GNU General Public License 2.0 or later, see https://www.gnu.org/licenses/old-licenses/gpl-2.0.en.html
 * @since   version 1.0
 */
class nodarkmode extends CMSPlugin
{
	const SUFFIX = 'nodarkmode';

	const COOKIE = [
		'name'     => 'reDimNoDarkmodeConfig',
		'validFor' => 31536000 // 1 year
	];

	/**
	 * @var WebAssetManager
	 * @since version 1.0
	 */
	private WebAssetManager $wa;

	/**
	 * @var array
	 * @since version 1.0
	 */
	private static array $files = [
		//Assetname //////////////// Dateipfad
		'template.atum.ltr'       => JPATH_ROOT . '/media/templates/administrator/atum/css/template.css',
		'webcomponent.joomla-tab' => JPATH_ROOT . '/media/templates/administrator/atum/css/vendor/joomla-custom-elements/joomla-tab.css',
		'choicesjs'               => JPATH_ROOT . '/media/templates/administrator/atum/css/vendor/choicesjs/choices.css',
	];

	/**
	 * @var array
	 * @since version 1.0
	 */
	private array $imported = [];

	/**
	 * @var array
	 * @since version 1.0
	 */
	private static array $forimport = [];

	private array $disabled = [];

	/**
	 * @var boolean
	 * @since version 1.0
	 */
	private bool $isCompleteReload;

	/**
	 * @var ?string
	 * @since version 1.0
	 */
	private ?string $currentstate; // Active / Inactive

	/**
	 * @var Uri
	 * @since version 1.0
	 */
	private Uri $uri;

	/**
	 * @var string
	 * @since version 1.0
	 */
	private string $manualstyles = '
	/* reDim NoDarkmode inline Style overrides */
	:root {
        --body-color: #212529 !important;
        --template-bg-dark-5 #fff !important;
        --link-color: #2a69b8;
        --link-color-rgb: rgb(42, 105, 184);
        }
        * {
            color-scheme: light !important;
        }
        .btn-close {
            filter: unset !important;
        }
        a {color: var(--template-link-color);}
        #jform_articletext {
            background: #fff !important;
            color: var(--body-color);
        }
        joomla-tab > div[role="tablist"] > button[role="tab"] {
            color: var(--primary);
        }
        joomla-tab > div[role="tablist"] > button[role="tab"][aria-expanded="true"] {
            background: var(--template-bg-dark-3);
            color: var(--primary);
		}
		.header-item-content.reDimNoDarkmodeBtn {
			padding: 0 !important;
		}
		.header-item-content.reDimNoDarkmodeBtn > a {
			padding: 0 !important;
		}
        ';

	public function __construct(
		DispatcherInterface $dispatcher,
		Uri                 $uri,
		array               $config
	)
	{
		$this->uri = $uri;
		parent::__construct($dispatcher, $config);
	}

	public function onBeforeCompileHead()
	{
		if ($this->getApplication()->isClient('site'))
		{
			return;
		}

		$this->currentstate     = $this->getConfigurationCookie() ?? 'active';
		$this->wa               = $this->getApplication()->getDocument()->getWebAssetManager();
		$this->isCompleteReload = !is_null($this->getApplication()->input->server->get('HTTP_PRAGMA'));

		$this->wa->registerAndUseScript(
			'nodarkmode.core',
			'plugins/system/nodarkmode/assets/js/noDarkmode.js'
		);

		static::makeImportList();

		$cmd = $this->uri->getVar('nodarkmode');

		switch ($cmd)
		{
			case 'toggle':
				$this->toggleMode();
				break;

			case null:
			default:
				break;
		}

		if ($this->currentstate !== 'inactive')
		{
			$this->execute();
		}

		if ($this->params->get('rendertoggler') || is_null($this->params->get('rendertoggler')))
		{
			$this->renderToggler();
		}
		unset($this->imported);
	}

	/**
	 * Automatically disables Assets based on their Name
	 *
	 * @since version 1.0
	 */
	private function intelligentAssetDeactivation()
	{
		$assets = $this->wa->getAssets('style');

		array_walk($assets, function (WebAssetItem $asset) {
			if (str_contains($asset->getName(), 'dark'))
			{
				$this->wa->disableAsset('style', $asset->getName());
				$this->disabled[] = $asset->getName();
			}
		});
	}

	/**
	 * Toggles the Plugin-Execution On/Off
	 *
	 * @return void
	 * @since version 1.0
	 */
	private function toggleMode(): void
	{
		switch ($this->currentstate)
		{
			case 'active':
				$this->currentstate = 'inactive';
				break;

			case 'inactive':
				$this->currentstate = 'active';
				break;

			default:
				$this->currentstate = ($this->params->get('active')) ? 'active' : 'inactive';
				break;
		}

		$query = $this->uri->getQuery(true);

		if (isset($query['nodarkmode']))
		{
			unset($query['nodarkmode']);
		}

		$query = http_build_query($query);

		$url = $this->uri->getHost();
		$url .= $this->uri->getPath();
		$url .= (!empty($query)) ? '?' . $query : '';

		static::updateConfigurationCookie($this->currentstate);

		header("Location: https://" . $url);
	}

	/**
	 * Executes the Core Plugin Functionality
	 *
	 * @return void
	 * @since version 1.0
	 */
	private function execute(): void
	{
		if ($this->isCompleteReload || !$this->allFilesCached() && $this->currentstate === 'active')
		{
			$this->grabCSS();
			$this->removeDarkmodeSelectors();
			$this->compressCSS();
			$this->cacheModifiedStyles();
		}

		if ($this->params->get('iad'))
		{
			$this->intelligentAssetDeactivation();
		}

		$this->excludeNativeJoomlaAssets();
		$this->includeModifiedStyles();
	}

	/**
	 * Updates the Configuration Cookie
	 *
	 * @param   string  $mode  The current Mode (active / inactive)
	 *
	 *
	 * @return void
	 * @since version 1.0
	 */
	private static function updateConfigurationCookie(string $mode): void
	{
		setcookie(self::COOKIE['name'], $mode, [
			'sameSite' => 'lax',
			'secure'   => false,
			'expires'  => (time() + self::COOKIE['validFor'])
		]);
	}

	/**
	 * Get the current Configuration Cookie
	 *
	 * @return ?string
	 *
	 * @since version 1.0
	 */
	private function getConfigurationCookie(): ?string
	{
		return $this->getApplication()->input->cookie->get(self::COOKIE['name']);
	}

	/**
	 * Checks, if all Files from the
	 * Importlist are currenty cached
	 *
	 * @return boolean
	 *
	 * @since version 1.0
	 */
	private function allFilesCached(): bool
	{
		foreach (static::$forimport as $paths)
		{
			if (!file_exists($paths->newfile))
			{
				return false;
			}
		}

		return true;
	}

	/**
	 * Grabs all selected CSS-Files and combines them into a string
	 *
	 * @return void
	 * @since version 1.0
	 */
	private function grabCSS(): void
	{
		foreach (static::$files as $asset => $file)
		{
			if (!is_readable($file) || !$this->wa->assetExists('style', $asset))
			{
				continue;
			}

			$this->imported[$file] = file_get_contents($file);
		}
	}

	/**
	 * Minifes/Compresses all imported CSS Data
	 *
	 * @link  https://stackoverflow.com/questions/15195750/minify-compress-css-with-regex
	 *
	 * @return void
	 * @since version 1.0
	 */
	private function compressCSS(): void
	{
		$comments = <<<'EOS'
		(?sx)
		    # don't change anything inside of quotes
		    ( "(?:[^"\\]++|\\.)*+" | '(?:[^'\\]++|\\.)*+' )
		|
		    # comments
		    /\* (?> .*? \*/ )
		EOS;

		$everything_else = <<<'EOS'
		(?six)
		    # don't change anything inside of quotes
		    ( "(?:[^"\\]++|\\.)*+" | '(?:[^'\\]++|\\.)*+' )
		|
		    # spaces before and after ; and }
		    \s*+ ; \s*+ ( } ) \s*+
		|
		    # all spaces around meta chars/operators (excluding + and -)
		    \s*+ ( [*$~^|]?+= | [{};,>~] | !important\b ) \s*+
		|
		    # all spaces around + and - (in selectors only!)
		    \s*([+-])\s*(?=[^}]*{)
		|
		    # spaces right of ( [ :
		    ( [[(:] ) \s++
		|
		    # spaces left of ) ]
		    \s++ ( [])] )
		|
		    # spaces left (and right) of : (but not in selectors)!
		    \s+(:)(?![^\}]*\{)
		|
		    # spaces at beginning/end of string
		    ^ \s++ | \s++ \z
		|
		    # double spaces to single
		    (\s)\s+
		EOS;

		$search_patterns  = ["%{$comments}%", "%{$everything_else}%"];
		$replace_patterns = ['$1', '$1$2$3$4$5$6$7$8'];

		foreach ($this->imported as &$data)
		{
			$data = preg_replace($search_patterns, $replace_patterns, $data);
		}
	}

	/**
	 * Deactivates all Darkmode-choosing Mediaqueries inside the CSS
	 *
	 * @return void
	 * @since version 1.0
	 */
	private function removeDarkmodeSelectors(): void
	{
		foreach ($this->imported as &$data)
		{
			$data = str_replace(
				'prefers-color-scheme: dark',
				'prefers-color-scheme: none',
				$data
			);
		}
	}

	/**
	 * Renders the Toggler Button
	 *
	 * @return void
	 * @since version 1.0
	 */
	private function renderToggler(): void
	{
		$query = $this->uri->getQuery();
		$path  = $this->uri->getPath();

		if (!str_contains($query, 'nodarkmode='))
		{
			$newQueryString = empty($query) ? 'nodarkmode=toggle' : $query . '&nodarkmode=toggle';
			$link           = $path . '?' . $newQueryString;
		}
		else
		{
			$link = $path . '?' . $query;
		}

		$disabledAssets = implode(', ', array_map(function ($asset) {
			return '"' . $asset . '"';
		}, $this->disabled));

		$this->wa->addInlineScript('
		const lang = new noDarkmode.Language("' . $this->getApplication()->getLanguage()->getTag() . '");
		
        if(noDarkmode.browserPrefersScheme("dark"))
        {
            document.addEventListener("DOMContentLoaded", function() {
            
                const toggler = new noDarkmode.Toggler("header-item-content");
                toggler.dom.href = "' . $link . '";
            
                const state = noDarkmode.getCookie("' . self::COOKIE['name'] . '");
                toggler.setState(state);
            });
            
            const disabled = [' . $disabledAssets . '];
            
            if(disabled.length !== 0)
            {
                console.info(lang._("ND_ASSETS_DEACTIVATED") + JSON.stringify(disabled, null, "  "));
            }
        }
        else
        {
            console.info(lang._("ND_NOT_BROWSER_ENABLED"));
        }
        ', [], [], ['nodarkmode.core']);

		$this->wa->addInlineStyle('
		.reDimnoDarkmodeToggler {padding:0px; margin-right: 3px;}
		');
	}

	/**
	 * Caches the Modified Styles
	 *
	 * @return void
	 * @since version 1.0
	 */
	private function cacheModifiedStyles(): void
	{
		foreach (static::$forimport as $paths)
		{
			#list($newfile, $oldfile) = $paths;

			$content = $this->imported[$paths->oldfile];
			File::write($paths->newfile, $content);
		}
	}

	/**
	 * Creates an Import List containing the old and
	 * new Filepaths and the new Asset Name
	 *
	 * @since version 1.0
	 */
	public static function makeImportList(): void
	{
		foreach (static::$files as $asset => $path)
		{
			$files          = new \stdClass();
			$newname        = File::stripExt(basename($path)) . '.' . self::SUFFIX . '.css';
			$files->newfile = dirname($path) . '/' . $newname;
			$files->oldfile = $path;

			static::$forimport[$asset . '.' . self::SUFFIX] = $files;
		}
	}

	/**
	 *
	 * @return array
	 *
	 * @since version 1.0
	 */
	public static function getImportList(): array
	{
		if (empty(static::$forimport))
		{
			static::makeImportList();
		}

		return static::$forimport;
	}

	/**
	 * Excludes native Joomla! assets
	 *
	 * @return void
	 * @since version 1.0
	 */
	private function excludeNativeJoomlaAssets(): void
	{
		foreach (array_keys(static::$files) as $asset)
		{
			$this->wa->disableAsset('style', $asset);
			$this->disabled[] = $asset;
		}
	}

	/**
	 * Includes the modified Styles
	 *
	 * @return void
	 * @since version 1.0
	 */
	private function includeModifiedStyles(): void
	{
		$this->wa->addInlineStyle($this->manualstyles);

		foreach (static::$forimport as $newasset => $paths)
		{
			$this->wa->registerAndUseStyle($newasset, static::makeWebPath($paths->newfile));
		}
	}

	/**
	 * Converts an absolute path to a Web-accessible path
	 *
	 * @param   string  $uri  The absolute URI
	 *
	 * @return string
	 *
	 * @since version 1.0
	 */
	private static function makeWebPath(string $uri): string
	{
		$toberemoved = explode('/', JPATH_ROOT);
		$newuri      = [];
		$catch       = false;

		foreach (explode('/', $uri) as $uripart)
		{
			if (!in_array($uripart, $toberemoved) && !$catch)
			{
				$catch = true;
			}

			if ($catch)
			{
				$newuri[] = $uripart;
			}
		}

		return implode('/', $newuri);
	}
}